<!doctype html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>Pyonir Demo Page</title>
</head>
<body>
<div style="text-align: center;">
    <h1>Welcome to Pyonir!</h1>
    <em>Seeing this page means your application doesn't have a frontend or content.</em>
</div>
</body>
</html>